import argparse
import numpy as np
import vtk
from vtk.numpy_interface import algorithms as algs
from vtk.numpy_interface import dataset_adapter as dsa

import matplotlib.pylab as plt
import scipy.interpolate as si


def transfinite(lower, upper, left, right):
    """Make a transfinite interpolation.

    http://en.wikipedia.org/wiki/Transfinite_interpolation

    Args:
        lower (list): List of (x, y) tuples describing the lower bound
        upper (list): List of (x, y) tuples describing the upper bound
        left  (list): List of (x, y) tuples describing the left bound
        right (list): List of (x, y) tuples describing the right bound

    Example input for the lower boundary:
        lower = [(0.0, 0.0), (0.1, 0.3),  (0.5, 0.4)]
    """

    lower = np.array(lower)
    upper = np.array(upper)
    left = np.array(left)
    right = np.array(right)

    # interpolate B-spline through data points
    # here, a linear interpolant is derived "k=1"
    # splprep returns:
    # tck ... tuple (t,c,k) containing the vector of knots,
    #         the B-spline coefficients, and the degree of the spline.
    #   u ... array of the parameters for each given point (knot)
    tck_left, u_left = si.splprep(left.T, s=0, k=1)
    tck_right, u_right = si.splprep(right.T, s=0, k=1)
    tck_lower, u_lower = si.splprep(lower.T, s=0, k=1)
    tck_upper, u_upper = si.splprep(upper.T, s=0, k=1)

    # evaluate function at any parameter "0<=t<=1"
    def eta_left(t):
        return np.array(si.splev(t, tck_left, der=0))

    def eta_right(t):
        return np.array(si.splev(t, tck_right, der=0))

    def xi_bottom(t):
        return np.array(si.splev(t, tck_lower, der=0))

    def xi_top(t):
        return np.array(si.splev(t, tck_upper, der=0))

    nodes = np.zeros((len(u_left) * len(u_lower), 2))

    # corner points
    c1 = xi_bottom(0.0)
    c2 = xi_top(0.0)
    c3 = xi_bottom(1.0)
    c4 = xi_top(1.0)

    for i, xi in enumerate(u_lower):
        xi_t = u_upper[i]
        for j, eta in enumerate(u_left):
            eta_r = u_right[j]

            node = i * len(u_left) + j

            # formula for the transinite interpolation
            point = (1.0 - xi) * eta_left(eta) + xi * eta_right(eta_r) + \
                (1.0 - eta) * xi_bottom(xi) + eta * xi_top(xi_t) - \
                ((1.0 - xi) * (1.0 - eta) * c1 + (1.0 - xi) * eta * c2 +
                 xi * (1.0 - eta) * c3 + xi * eta * c4)

            nodes[node, 0] = point[0]
            nodes[node, 1] = point[1]

    return nodes


def create_spline(x_array, y_array, z_array, resolution):

    coords = algs.make_vector(x_array, y_array, z_array)
    pts = vtk.vtkPoints()
    pts.SetData(dsa.numpyTovtkDataArray(coords, 'Points'))
    spline = vtk.vtkParametricSpline()
    spline.SetPoints(pts)
    spline_source = vtk.vtkParametricFunctionSource()
    spline_source.SetParametricFunction(spline)
    spline_source.SetUResolution(resolution)
    spline_source.SetVResolution(resolution)
    spline_source.SetWResolution(resolution)
    spline_source.Update()

    return spline_source.GetOutput()


def main():

    parser = argparse.ArgumentParser()

    parser.add_argument('--polyline1', '-p1', default='poly1.csv')
    parser.add_argument('--polyline2', '-p2', default='poly2.csv')
    args = parser.parse_args()

    resolution = 200

    # Create a polyline on side 1 of blade
    p1_data = np.genfromtxt(args.polyline1, dtype=None, names=True, delimiter=',', autostrip=True, encoding='utf-8')
    p1_sp = create_spline(p1_data['Points_0'][::-1], p1_data['Points_1'][::-1], p1_data['Points_2'][::-1], resolution=resolution)

    p1_writer = vtk.vtkXMLPolyDataWriter()
    p1_writer.SetFileName('p1.vtp')
    p1_writer.SetInputData(p1_sp)
    p1_writer.Write()

    # Create a polyline on side 2 of blade
    p2_data = np.genfromtxt(args.polyline2, dtype=None, names=True, delimiter=',', autostrip=True, encoding='utf-8')
    p2_sp = create_spline(p2_data['Points_0'], p2_data['Points_1'], p2_data['Points_2'], resolution=resolution)

    p2_writer = vtk.vtkXMLPolyDataWriter()
    p2_writer.SetFileName('p2.vtp')
    p2_writer.SetInputData(p2_sp)
    p2_writer.Write()

    # Outer endpoint of polyline 1
    p3_pts = vtk.vtkPoints()
    x31 = p1_data['Points_0'][0]
    y31 = p1_data['Points_1'][0]
    z31 = p1_data['Points_2'][0]
    p31 = np.array([x31, z31])
    norm3 = np.linalg.norm(p31)
    p31 = p31 / norm3

    # Angle of the outer endpoint 1 with x-axis
    angle1 = np.arcsin(p31[1])
    print(angle1*180.0/np.pi)

    # Outer end point of polyline 2
    x32 = p2_data['Points_0'][-1]
    y32 = p2_data['Points_1'][-1]
    z32 = p2_data['Points_2'][-1]
    p32 = np.array([x32, z32])
    p32 = p32 / np.linalg.norm(p32)

    # Angle of the outer endpoint 2 with x-axis
    angle2 = np.arcsin(p32[1])
    print(angle2*180.0/np.pi)

    # Create a midpoint between the outer endpoints
    mid_angle = (angle2 - angle1) * 0.5
    print('mid_angle', mid_angle*180.0/np.pi)

    x33 = x31 * np.cos(mid_angle) - z31 * np.sin(mid_angle)
    y33 = y31
    z33 = x31 * np.sin(mid_angle) + z31 * np.cos(mid_angle)

    print(x31,y31,z31)
    print(x32,y32,z32)
    print(x33,y33,z33)

    # Inner end point of polyline 1
    p4_pts = vtk.vtkPoints()
    x41 = p1_data['Points_0'][-1]
    y41 = p1_data['Points_1'][-1]
    z41 = p1_data['Points_2'][-1]
    p41 = np.array([x41, z41])
    norm4 = np.linalg.norm(p41)
    p41 = p41 / norm4

    angle1 = np.arccos(p41[0])
    print('angle1: ', angle1*180.0/np.pi)

    # Inner end point of polyline 2
    x42 = p2_data['Points_0'][0]
    y42 = p2_data['Points_1'][0]
    z42 = p2_data['Points_2'][0]
    p42 = np.array([x42, z42])
    p42 = p42 / np.linalg.norm(p42)

    print('norm3: ', np.linalg.norm(p41), np.linalg.norm(p42))

    angle2 = np.arccos(p42[0])
    print('angle2: ', angle2*180.0/np.pi)

    # Create a midpoint between the inner endpoints
    mid_angle = (angle2 - angle1) * 0.5
    print('mid_angle', mid_angle*180.0/np.pi)

    x43 = x41 * np.cos(mid_angle) - z41 * np.sin(mid_angle)
    y43 = y41
    z43 = x41 * np.sin(mid_angle) + z41 * np.cos(mid_angle)

    print(x41,y41,z41)
    print(x42,y42,z42)
    print(x43,y43,z43)

    upper = [(x31,z31), (x33,z33), (x32,z32)]
    lower = [(x41,z41), (x43,z43), (x42,z42)]

    right = []
    for i in range(p1_sp.GetNumberOfPoints()):
        pnt = p1_sp.GetPoint(i)
        right.append((pnt[0], pnt[2]))

    left = []
    for i in range(p2_sp.GetNumberOfPoints()):
        pnt = p2_sp.GetPoint(i)
        left.append((pnt[0], pnt[2]))

    nodes = transfinite(lower, upper, left, right)
    print(nodes)

    lower = np.array(lower)
    upper = np.array(upper)
    left = np.array(left)
    right = np.array(right)

    plt.plot(lower[:, 0], lower[:, 1], 'r', marker='o', clip_on=False)
    plt.plot(upper[:, 0], upper[:, 1], 'b', marker='o', clip_on=False)
    plt.plot(left[:, 0], left[:, 1], 'g', marker='o', clip_on=False)
    plt.plot(right[:, 0], right[:, 1], 'k', marker='o', clip_on=False)
    plt.title('Outline of a block')
    plt.show()

    vlines = list()
    vline = list()
    for i, node in enumerate(nodes):
        vline.append(node.tolist())
        if (i+1) % (resolution+1) == 0:
            vlines.append(vline)
            vline = []

    ulines = []
    uline = []
    for i in range(len(vlines[0])):
        for vline in vlines:
            uline.append(vline[i])
        ulines.append(uline)
        uline = []

    for vline in vlines:
        plt.plot(np.array(vline)[:, 0], np.array(vline)[:, 1], 'm')

    for uline in ulines:
        plt.plot(np.array(uline)[:, 0], np.array(uline)[:, 1], 'm')

    plt.plot(nodes[:, 0], nodes[:, 1], 'mo', zorder=10, clip_on=False)
    plt.title('Points created by TFI')

    plt.show()

    # ts_001.vtp will be the midline between blade lines
    for i, vline in enumerate(vlines):
        vline = np.array(vline).reshape(-1, 2)
        coords = algs.make_vector(vline[:,0], np.full(resolution+1, y31), vline[:,1])
        pts = vtk.vtkPoints()
        pts.SetData(dsa.numpyTovtkDataArray(coords, 'Points'))
        spline = vtk.vtkParametricSpline()
        spline.SetPoints(pts)
        spline_source = vtk.vtkParametricFunctionSource()
        spline_source.SetParametricFunction(spline)
        spline_source.SetUResolution(resolution)
        spline_source.SetVResolution(resolution)
        spline_source.SetWResolution(resolution)
        spline_source.Update()

        writer = vtk.vtkXMLPolyDataWriter()
        writer.SetFileName(f'ts_{i:0>3}.vtp')
        writer.SetInputData(spline_source.GetOutput())
        writer.Write()


if __name__ == '__main__':
    main()
